const express = require('express')
const SchedulerRepo = require('../repository/scheduler')
const AccountRepo = require('../repository/account')

// create router
// params
// - db <connection>: instance of mongoose connection
// return express router
module.exports = function (db, oauth) {
  const schedulerRepo = new SchedulerRepo(db)
  const accountRepo = new AccountRepo(db, oauth)
  var router = express.Router()

  router.route('/item')
  .post(function (req, res) {
    accountRepo.authenticate(req, function (authErr, user) {
      if (authErr) {
        res.status(401).json({
          message: 'authenticate error',
          data: authErr
        })
        return
      }

      req.body.author = user.user._id

      schedulerRepo.create(req.body, function (insertErr) {
        if (insertErr) {
          res.status(500).json({
            message: 'internal error',
            data: insertErr
          })
          return
        }
        res.json({message: 'done'})
      })
    })
  })

  router.route('/')
  .get(function (req, res) {
    var selector = {
      pageSize: req.query.pageSize || 16,
      pageIndex: req.query.pageIndex || 1
    }

    schedulerRepo.find(selector, function (error, items) {
      if (error) {
        res.status(500).json(error)
        return
      }
      if (items.length === 0) {
        res.status(404).json({
          message: 'not found scheduler'
        })
        return
      }
      res.json(items)
    })
  })

  router.route('/item/:id')
  .get(function (req, res) {
    schedulerRepo.findById(req.params.id, function (err, item) {
      if (err) {
        res.json(err)
        return
      }

      res.json(item)
    })
  })
  .delete(function (req, res) {
    accountRepo.authenticate(req, function (authErr, user) {
      if (authErr) {
        res.status(401).json({
          message: 'authenticate error',
          data: authErr
        })
        return
      }

      schedulerRepo.remove(req.params.id, user.user._id, function (removeErr, removeRes) {
        if (removeErr || removeRes.result.n === 0) {
          res.status(400).json({
            message: 'can not remove',
            data: removeErr
          })
          return
        }
        res.json({
          message: 'done'
        })
      })
    })
  })

  router.route('/search/:keyword/size/:pageSize/index/:pageIndex')
  .get(function (req, res) {
    var selector = {
      pageIndex: Number(req.params.pageIndex),
      pageSize: Number(req.params.pageSize)
    }
    schedulerRepo.search(req.params.keyword, selector, function (error, items) {
      if (error) {
        res.json(error)
        return
      }

      res.json(items)
    })
  })

  router.route('/search/me/:keyword/size/:pageSize/index/:pageIndex')
  .get(function (req, res) {
    accountRepo.authenticate(req, function (authErr, user) {
      if (authErr || !user.user) {
        res.status(401).json({
          message: 'authenticate error',
          data: authErr
        })
        return
      }

      var selector = {
        pageIndex: Number(req.params.pageIndex),
        pageSize: Number(req.params.pageSize)
      }
      schedulerRepo.searchAuthor(user.user._id, user.user.schedulers, req.params.keyword, selector, function (searchErr, items) {
        if (searchErr) {
          res.json(searchErr)
          return
        }
        res.json(items)
      })
    })
  })

  return router
}
