const googleapis = require('googleapis')

// account repository
// using to query account in database and other oauth provider
// params
//  - db <connection>: mongoose database connection
//  - oauth <object>: contains information about oauth provider
// return instance of account repository
function AccountRepository (db) {
  this._db = db
}

// shortcut to account repository prototype
// using to access to repository prototype more quickly
var prototype = AccountRepository.prototype

// authenticate user with external oauth provider
// params
//  - request <object>: express request object contain information for request
//  - callback <function>: function will call
//    params
//     - error <object>: error during process
//     - user <object>: contains user properties
prototype.authenticateEx = function (request, callback) {
  var self = this
  var plus = googleapis.plus('v1')
  var credentials = {
    access_token: request.headers['authorization'],
    refresh_token: request.headers['refresh_token']
  }

  self.oauth2.setCredentials(credentials)
  plus.people.get({userId: 'me', auth: self.oauth2}, function (error, user) {
    callback(error, user)
  })
}

// authenticate user and get account information in database
// params
//  - request <object>: express request object contain information for request
//  - callback <function>: function will call
//    params
//     - error <object>: error during process
//     - data <object>: contains user external and user object
prototype.authenticate = function (request, callback) {
  var self = this
  self.authenticateEx(request, function (errEx, userEx) {
    if (errEx) {
      callback(errEx)
      return
    }

    self.model.findOne({ subject: userEx.id }, function (err, user) {
      if (err) {
        callback(err)
        return
      }
      callback(err, {userex: userEx, user: user})
    })
  })
}

// find user by pair
// params
//  - issuser <string>: issuser property in oauth context. if not specific,
//    only subject will be search
//  - subject <string>: subject property in oauth context
//  - callback <function>: function will call
//    params
//       - error <object>: error return by process
//       - item <object>: account if found
prototype.findByIssuer = function (issuer, subject, callback) {
  var self = this
  var filter = {subject: subject}
  if (issuer) {
    filter.issuer = issuer
  }
  self.model.findOne(filter, function (err, item) {
    callback(err, item)
  })
}

// create new user into database
// params
//  - item <object>: account to create
//  - callback <function>: function will called
//    params
//       - error <object>: error return by process
//       - item <object>: account created if success
prototype.create = function (item, callback) {
  var self = this

  // check account is exist with pair
  //  - issuer
  //  - subject
  var filter = {issuer: item.issuer, subject: item.subject}
  self.model.findOne(filter, function (error, item_) {
    if (error) {
      callback(error)
      return
    }
    if (item_) {
      callback({message: 'account is exists'})
      return
    }

    // create new account
    self.model.create(item, function (error_, item__) {
      callback(error_, item__)
    })
  })
}

// update scheduler which user using
// params
//  - subject <string>: identity of account in provider
//  - schedulerId <string>: scheduler identity
//  - callback <function>: function will call
//    params
//     - error <object>: error during process
prototype.updateScheduler = function (subject, schedulerId, callback) {
  const self = this

  // identify account
  this.model.findOne({ subject: subject }, function (errAcc, account) {
    if (errAcc) {
      callback(errAcc)
      return
    }

    // verify that scheduler is exists
    var schedulerRepo = new SchedulerRepo(self.db)
    schedulerRepo.model.findById(schedulerId, function (errScheduler, scheduler) {
      if (errScheduler) {
        callback(errScheduler)
        return
      }

      // update scheduler property
      if (account.schedulers.indexOf(schedulerId) < 0) {
        account.schedulers.push(schedulerId)
      }
      account.scheduler = schedulerId
      account.save()

      // success
      callback()
    })
  })
}

// update user language
// params
//  - subject <object>: identity of account in provider
//  - language <string>: language code, same as 'eng'
//  - callback <function>: function will called
//    params
//       - error <object>: error return by process
prototype.updateLanguage = function (subject, language, callback) {
  this.model.findOne({ subject: subject }, function (errAcc, account) {
    if (errAcc) {
      callback(errAcc)
      return
    }

    account.language = language
    account.save(callback)
  })
}

// export module
module.exports = AccountRepository
